/*++

Copyright (c) Microsoft Corporation.  All rights reserved.

    THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
    KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
    PURPOSE.

Module Name:

    tracectl.c

Environment:

    User mode Win32 console application

Revision History:


--*/
#include <windows.h>
#include <winioctl.h>
#include <tchar.h>
#include <stdio.h>
#include "install.h"
#include "cpuid.h"
#include <conio.h>

#define WAIT_TIME       10

BOOLEAN
SetupDriverName(
    LPTSTR DriverLocation
    );

int _cdecl _tmain(int argc, TCHAR* argv[])
{
  DWORD  dwError = ERROR_SUCCESS;
  TCHAR  driverLocation[MAX_PATH];
  DWORD  dwOutBuffer[2048];
  DWORD  dwOutBufferCount ;
  int ch;
  
	if(argc != 2) {
		_tprintf(_T("usage: fastsyscall.exe [-i|-r]\n"));
		return 0;
	}
  
	if(CPUID_Available()) {
		_tprintf(_T("cpuid is available\n"));
	} else {
		_tprintf(_T("your cpu does not support cpuid\n"));
		return 1;
	}
	
	if(SYSCALL_Available()) {
		_tprintf(_T("cpu supports syscall\n"));
	}
	if(SYSENTER_Available()) {
		_tprintf(_T("cpu supports sysenter\n"));
	}
	
	if(!SYSCALL_Available() && !SYSENTER_Available()) {
		_tprintf(_T("cpu does not support any fast system call\n"));
		return 1;
	}
        //
        //  Setup full path to driver name
        //
	if (!SetupDriverName(driverLocation)) {
		_tprintf(_T("%s.sys is not found\n"), DRIVER_NAME);
		return 2;
	}
	
	if(_tcscmp(_T("-i"), argv[1]) == 0) {
        //
        // Install driver
        //

        if (!ManageDriver(DRIVER_NAME,
                          driverLocation,
                          DRIVER_FUNC_INSTALL
                          )) {

            _tprintf(_T("Unable to install driver. \n"));

            //
            // Error - remove driver.
            //

            ManageDriver(DRIVER_NAME,
                         driverLocation,
                         DRIVER_FUNC_REMOVE
                         );
            
            return 3;
        }
		
		_tprintf(_T("Driver '%s' is installed\n"), DRIVER_NAME);
		return 0;
	} else if(_tcscmp(_T("-r"), argv[1]) == 0) {
		//
		//  stop the driver
		//
    
		ManageDriver(DRIVER_NAME,
                 driverLocation,
                 DRIVER_FUNC_REMOVE
                 );
		_tprintf(_T("Driver '%s' is removed\n"), DRIVER_NAME);

		return 0;
	} else {
		_tprintf(_T("usage: fastsyscall.exe [-i|-r]\n"));
		return 0;
	}
}

BOOLEAN
SetupDriverName(
    LPTSTR DriverLocation
    )
{
    HANDLE fileHandle;

    DWORD driverLocLen = 0;

    //
    // Get the current directory.
    //

    driverLocLen = GetCurrentDirectory(MAX_PATH,
                                       DriverLocation
                                       );

    if (!driverLocLen) {

        _tprintf(_T("GetCurrentDirectory failed!  Error = %d \n"), GetLastError());

        return FALSE;
    }

    //
    // Setup path name to driver file.
    //

    _tcscat(DriverLocation, _T("\\"));
    _tcscat(DriverLocation, DRIVER_NAME);
    _tcscat(DriverLocation, _T(".sys"));

    //
    // Insure driver file is in the specified directory.
    //

    if ((fileHandle = CreateFile(DriverLocation,
                                 GENERIC_READ,
                                 0,
                                 NULL,
                                 OPEN_EXISTING,
                                 FILE_ATTRIBUTE_NORMAL,
                                 NULL
                                 )) == INVALID_HANDLE_VALUE) {


        _tprintf(_T("Driver: '%s' is not in the current directory. \n"), DRIVER_NAME);

        //
        // Indicate failure.
        //

        return FALSE;
    }

    //
    // Close open file handle.
    //

    if (fileHandle) {

        CloseHandle(fileHandle);
    }

    //
    // Indicate success.
    //

    return TRUE;


}   // SetupDriverName

