/*
  Windows 2000 XP API Wrapper Pack
  Copyright (C) 2008 OldCigarette

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <ntddk.h>
#include <windef.h>
#include "msr.h"
#include "cpuid.h"

#define NAKED __declspec(naked)
#define ALIGN(n) __declspec(align(n))

//John voted for John

NTSTATUS DriverEntry(PDRIVER_OBJECT  DriverObject, PUNICODE_STRING RegistryPath);
VOID GetIDTR(PIDTR idtr);
PINT_GATE_DESC GetIntGate(DWORD index);
VOID RunEachProcessor(PKDPC Dpc, PVOID DeferredContext, PVOID SystemArgument1, PVOID SystemArgument2);
VOID Enable_SYSCALL();
VOID Enable_SYSENTER();
VOID Disable_SYSCALL();
VOID Disable_SYSENTER();
VOID ReplaceKiSystemService();
VOID RestoreKiSystemService();
VOID DriverUnload(PDRIVER_OBJECT DriverObject);

#ifdef ALLOC_PRAGMA
#pragma alloc_text(INIT, DriverEntry)
#pragma alloc_text(PAGE, DriverUnload)
#endif

KDPC dpc;
PVOID KiSystemService;
PVOID KiSystemService_Hook;
KEVENT DriverLock;

/*
Find me from NtQuerySystemInformation
#define KF_FAST_SYSCALL 0x00001000 (use sysexit)
.data:0047829C KeFeatureBits   dd 0                    ; DATA XREF: sub_4329DC+27r
.data:0047829C                                         ; sub_432AA0+1Fr ...
*/
#define KF_FAST_SYSCALL 0x00001000
#define KeFeatureBits *((PDWORD)0x8047829C)

#define FASTSYSCALL_ENABLE  ((PVOID)1)
#define FASTSYSCALL_DISABLE ((PVOID)2)

#define ENABLE_SYSCALL  1
#define ENABLE_SYSENTER 2
#define ENABLE_SYSEXIT  4
DWORD Flags;
BOOL  SYSCALL_Enabled;
BOOL  IntHook_Enabled;

VOID GetIDTR(PIDTR idtr) {
	__asm {
		mov  eax, idtr
		sidt [eax]
	}
}

PINT_GATE_DESC GetIntGate(DWORD index) {
	IDTR idtr;
	GetIDTR(&idtr);
	return &(((PINT_GATE_DESC)idtr.base)[index]);
}

VOID NAKED ALIGN(16) SYSCALL_Patch() {
	__asm {
		//only patch user mode code
		cmp  eax, 0f8h
		jae  fromkmode
		test dword ptr [esp+8h],  20000h //V86 mode
		jnz  fromkmode
		cmp  dword ptr [esp+4h],  SEG_SELECTOR(KGDT_R3_CODE, GLOBAL_DESC_TBL, RING_3)
		jnz  fromkmode
		cmp  dword ptr [esp+10h], SEG_SELECTOR(KGDT_R3_DATA, GLOBAL_DESC_TBL, RING_3)
		jnz  fromkmode
		
		//get eip, sub to int2e
		push ecx
		push eax
		mov  ecx, [esp+8]
		sub  ecx, 2
		mov  eax, ecx
		
		//lookup pte
		shr  ecx, 12
		shl  ecx, 2
		or   ecx, 0c0000000h

		//test write enabled (don't patch)
		test dword ptr [ecx], 2
		jnz  alldone

		//enable writing & patch
		or   dword ptr [ecx], 2
		invlpg [eax]
		or   dword ptr [ecx+4], 2
		invlpg [eax+1]
		
		mov  word ptr [eax], 050fh
		
		and  dword ptr [ecx], ~2
		invlpg [eax]
		and  dword ptr [ecx+4], ~2
		invlpg [eax+1]
/*		
		jmp  alldone
	writeenabled:
		mov  word ptr [eax], 050fh*/
	alldone:
		pop  eax
		pop  ecx
	fromkmode:
		jmp  dword ptr [KiSystemService]
	}
}

VOID NAKED ALIGN(16) SYSCALL_Entry() {
	__asm {
#if 1
		mov sp, fs
		mov ds, sp
		mov esp, 00000030h
		mov fs,  sp
		mov esp, fs:[40h]
#else
		mov esp, ss:[40h]
#endif
		mov esp, ss:[esp + 4]
		
		push SEG_SELECTOR(KGDT_R3_DATA, GLOBAL_DESC_TBL, RING_3) //return ss
		push edx                                                 //return esp
		sub  dword ptr ss:[esp], 4
		pushfd                                                   //return eflags
		or   dword ptr ss:[esp], 1<<9
		push SEG_SELECTOR(KGDT_R3_CODE, GLOBAL_DESC_TBL, RING_3) //return cs
		push ecx                                                 //return eip
		
		mov cx, ds
		mov fs, cx
		mov cx, SEG_SELECTOR(KGDT_R3_DATA, GLOBAL_DESC_TBL, RING_3)
		mov ds, cx
		
		pushfd
		and dword ptr ss:[esp], 0fffcbcffh
		popfd
		
		jmp dword ptr [KiSystemService]
	}
}

VOID ReplaceKiSystemService() {
	PINT_GATE_DESC desc;
	__asm cli
	desc = GetIntGate(0x2e);
	desc->addrlo = (WORD)(((DWORD)KiSystemService_Hook) & 0xFFFF);
	desc->addrhi = (WORD)(((DWORD)KiSystemService_Hook) >> 16);
	__asm sti
}

VOID RestoreKiSystemService() {
	PINT_GATE_DESC desc;
	__asm cli
	desc = GetIntGate(0x2e);
	desc->addrlo = (WORD)(((DWORD)KiSystemService) & 0xFFFF);
	desc->addrhi = (WORD)(((DWORD)KiSystemService) >> 16);
	__asm sti
}

VOID Disable_SYSCALL() {
	LARGE_INTEGER msr;
	msr.QuadPart = __readmsr(0xC0000080).QuadPart & ~0x1LL;
	__writemsr(0xC0000080, msr);
}

VOID Disable_SYSENTER() {
	LARGE_INTEGER msr;
	msr.HighPart = 0;
	msr.LowPart = 0;
	__writemsr(0x174, msr);
}

VOID Enable_SYSCALL() {
	LARGE_INTEGER msr;
	STAR star;
	
	star.eip     = (DWORD)SYSCALL_Entry;
	star.call_cs = SEG_SELECTOR(KGDT_R0_CODE, GLOBAL_DESC_TBL, RING_0);
	star.ret_cs  = SEG_SELECTOR(KGDT_R3_CODE, GLOBAL_DESC_TBL, RING_3);
	__writemsr(0xC0000081, *(LARGE_INTEGER *)&star);
	
	msr.QuadPart = __readmsr(0xC0000080).QuadPart | 0x1LL;
	__writemsr(0xC0000080, msr);
}

VOID Enable_SYSENTER() {
	LARGE_INTEGER msr;

	msr.HighPart = 0;
	msr.LowPart = SEG_SELECTOR(KGDT_R0_CODE, GLOBAL_DESC_TBL, RING_0);
	__writemsr(0x174, msr);
	
	msr.LowPart = 0;
	__writemsr(0x175, msr);
	
	msr.LowPart = (DWORD)SYSCALL_Entry;
	__writemsr(0x176, msr);
}

VOID RunEachProcessor(PKDPC Dpc, PVOID DeferredContext, PVOID mode, PVOID cpu) {
	CHAR i;
	i = (CHAR)cpu + 1;
	
	if(mode == FASTSYSCALL_ENABLE) {
		if(!SYSCALL_Enabled) {
			if(Flags&ENABLE_SYSCALL) Enable_SYSCALL();
			if(Flags&ENABLE_SYSENTER || Flags&ENABLE_SYSEXIT) Enable_SYSENTER();
			KdPrint(("fastsyscall: fast sys call enabled on cpu %d\n", (DWORD)i));
			if(i == *KeNumberProcessors) {
				SYSCALL_Enabled = TRUE;
				i = 0;
			}
		} else if(Flags&ENABLE_SYSCALL) {
			ReplaceKiSystemService();
			KdPrint(("fastsyscall: IDT patched on cpu %d\n", (DWORD)i));
			if(i == *KeNumberProcessors) IntHook_Enabled = TRUE;
		}
	} else if(mode == FASTSYSCALL_DISABLE) {
		if(IntHook_Enabled) {
			RestoreKiSystemService();
			if(i == *KeNumberProcessors) {
				IntHook_Enabled = FALSE;
				i = 0;
			}
		} else {
			if(Flags&ENABLE_SYSCALL) Disable_SYSCALL();
			if(Flags&ENABLE_SYSENTER || Flags&ENABLE_SYSEXIT) Disable_SYSENTER();
			if(i == *KeNumberProcessors) SYSCALL_Enabled = FALSE;
		}
	}
	
	if(i < *KeNumberProcessors) {
		KeSetTargetProcessorDpc(&dpc, i);
		KeInsertQueueDpc(&dpc, mode, (PVOID)i);
	} else {
		//done processing
		if(mode == FASTSYSCALL_ENABLE && Flags & ENABLE_SYSEXIT)
			KeFeatureBits |= KF_FAST_SYSCALL;
		if(mode == FASTSYSCALL_DISABLE)
			KeSetEvent(&DriverLock, 0, FALSE);
	}
}

VOID DriverUnload(PDRIVER_OBJECT DriverObject) {
	//Unloading the driver will crash all open apps which have made a sys call
	if(Flags & ENABLE_SYSEXIT)
		KeFeatureBits &= ~KF_FAST_SYSCALL;
	KeSetTargetProcessorDpc(&dpc, 0);
	KeInsertQueueDpc(&dpc, FASTSYSCALL_DISABLE, (PVOID)0);
	
	KeWaitForSingleObject(&DriverLock, Executive, KernelMode, FALSE, NULL);
}

NTSTATUS DriverEntry(PDRIVER_OBJECT DriverObject, PUNICODE_STRING RegistryPath) {
	PINT_GATE_DESC desc;
	NTSTATUS status;

	KdPrint(("fastsyscall: DriverEntry\n"));
	
	if(!CPUID_Available()) {
		KdPrint(("fastsyscall: cpuid not available"));
		return STATUS_DEVICE_NOT_READY;
	}
	
	Flags = 0;
	if(SYSENTER_Available()) {
		KdPrint(("fastsyscall: SYSENTER detected"));
		Flags |= ENABLE_SYSEXIT;
	}
	if(SYSCALL_Available()) {
		KdPrint(("fastsyscall: SYSCALL detected"));
		Flags |= ENABLE_SYSCALL;
	}
	if(!Flags) {
		KdPrint(("fastsyscall: no fast system call available"));
		return STATUS_DEVICE_NOT_READY;
	}
	
	DriverObject->DriverUnload = DriverUnload;
	
	desc = GetIntGate(0x2e);
	KiSystemService = (PVOID)(desc->addrlo | (desc->addrhi<<16));
	KiSystemService_Hook = SYSCALL_Patch;
	
	SYSCALL_Enabled = IntHook_Enabled = FALSE;
	
	KeInitializeEvent(&DriverLock, NotificationEvent, FALSE);
	
	KeInitializeDpc(&dpc, RunEachProcessor, NULL);
	KeSetTargetProcessorDpc(&dpc, 0);
	KeInsertQueueDpc(&dpc, FASTSYSCALL_ENABLE, (PVOID)0);
		
	return STATUS_SUCCESS;
}
