/*
  Windows 2000 XP API Wrapper Pack
  Copyright (C) 2008 OldCigarette

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <windows.h>
#include <stdio.h>
#include "win2k_xp_debug.h"

HANDLE hLogFile;
HANDLE hMutex;
int debug;

WORD DebugColors[8] = {
	0,
	FOREGROUND_RED | FOREGROUND_INTENSITY,
	FOREGROUND_GREEN | FOREGROUND_INTENSITY,
	FOREGROUND_GREEN | FOREGROUND_RED | FOREGROUND_INTENSITY,
	FOREGROUND_BLUE | FOREGROUND_INTENSITY,
	FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_INTENSITY,
	FOREGROUND_BLUE | FOREGROUND_RED | FOREGROUND_INTENSITY,
	FOREGROUND_RED | FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_INTENSITY
};

WORD DefaultColor = FOREGROUND_RED | FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_INTENSITY;

void __cdecl vDbgPrintf(int level, LPCSTR text, va_list va) {
	static char tstr[2048];
	char *p, *s;
    DWORD bw;
    HANDLE hCon;
	
	if(level >= debug) return;
	
	hCon = GetStdHandle(STD_OUTPUT_HANDLE);
    if(hCon == INVALID_HANDLE_VALUE) return;
	
	//Yeah don't be debugging in any of the used console functions!!!!!
	WaitForSingleObject(hMutex, INFINITE);
		_vsnprintf(tstr, sizeof(tstr), text, va);
		
		OutputDebugString(tstr);
		
		SetConsoleTextAttribute(hCon, DefaultColor);
		for(s = p = tstr; *p != '\0'; p++) {
			if(*p == '^' && *(p+1) != '\0') {
				if(*(p+1) != '^') {
					WriteFile(hCon, s, p - s, &bw, NULL);
					SetConsoleTextAttribute(hCon, DebugColors[*(p+1) & 7]);
				} else {
					WriteFile(hCon, s, p - s + 1, &bw, NULL);
				}
				s = ++p + 1;
			}
		}
		WriteFile(hCon, s, p - s, &bw, NULL);
		if(hLogFile != INVALID_HANDLE_VALUE)
			WriteFile(hLogFile, s, p - s, &bw, NULL);
		SetConsoleTextAttribute(hCon, DefaultColor);
	ReleaseMutex(hMutex);
}

void __cdecl DbgPrintf(int level, LPCSTR text, ...) {
	va_list argptr;
	va_start(argptr, text);
		vDbgPrintf(level, text, argptr);
	va_end(argptr);
}

int __cdecl DebugLevel() {
	return debug;
}

void __cdecl InitDebug() {
	hMutex = CreateMutex(NULL, FALSE, NULL);
	
	//if(GetStdHandle(STD_OUTPUT_HANDLE) == NULL) {
		AllocConsole();
		SetConsoleTitle("Win2k XP Wrapper Debug Console");
		SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_RED);
	//}
	hLogFile = CreateFile("win2kxp_log.txt", GENERIC_READ | GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
    
    DbgPrintf(DBG_ALWAYS,
		S_YELLOW
		"\nDebug Console Startup\n"
		  "---------------------\n");
}
