/*
  Windows 2000 XP API Wrapper Pack
  Copyright (C) 2008 OldCigarette

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#define WIN2K_XP_DEBUG_STATIC

#include <stdio.h>
#include "win2k_xp_kernel32.h"
#include "../common/win2k_xp_config.h"
#include "../common/win2k_xp_debug.h"
#include "../common/detours.h"

/*Global Data*/
DWORD RtlDispatchException_resume;
ExceptionVectorList_t ExceptionVectorList;

OSVERSIONINFOEXA versionInfoA;
OSVERSIONINFOEXW versionInfoW;

BOOL GetModuleOriginal = FALSE;
HMODULE thisModule = NULL;
HMODULE hookedModule = NULL;

GEOID geoRegion;
GEOID geoNation;

WCHAR systemDirectoryW[MAX_PATH];
CHAR  systemDirectoryA[MAX_PATH];

RTL_CRITICAL_SECTION dllDirectoryLock;
WCHAR dllDirectoryW[MAX_PATH];

//GetGeoInfo
WCHAR geoISO2W[3];
WCHAR geoISO3W[4];
WCHAR geoRFC1766W[16];
WCHAR geoFriendlyW[128];
WCHAR geoOfficialW[128];

BOOL RewriteExportTable = FALSE;
BOOL verifyVersion = FALSE;

INT     numFilterModules = 0;
INT     versionFaking = VERSION_FAKING_EXCLUDE;
CHAR   *filterModuleNames[128];
HMODULE filterModules[128];

#define MAX_ENV 32767

//This is an internal function that needs to be patched, right now
//it is not because I don't think this address is very reliable

//7C5AE525   E8 73120000      CALL KERNEL_1.7C5AF79D
#define GETLOCALECOMPARE_ADDY 0x7C5AF79D
GetLocaleCompare_t GetLocaleCompare;

DWORD WINAPI XP_GetLocaleCompare(LCID locale, DWORD dwCmpFlags) {
	if(locale == LOCALE_INVARIANT) return GetLocaleCompare(LOCALE_SYSTEM_DEFAULT, dwCmpFlags);
	else return GetLocaleCompare(locale, dwCmpFlags);
}

int snwprintf(wchar_t *buffer, size_t count, const wchar_t *format, ...) {
	int i;
	va_list argptr;
	va_start(argptr, format);
	i = _vsnwprintf(buffer, count, format, argptr);
	va_end(argptr);
	return i;
}
  
DWORD GetStringA(DWORD len, LPSTR dest, LPCSTR source) {
	DWORD c;
	PCSTR p;

	for(c = 0, p = source; c < len && *p != '\0'; c++, p++) {
		dest[c] = *p;
	}

	if(c < len && *p == '\0') {
		dest[c] = '\0';
		return c;
	}

	while(*p) { c++; p++; }
	return c + 1;
}

DWORD GetStringW(DWORD len, LPWSTR dest, LPCWSTR source) {
	DWORD c;
	PCWSTR p;

	for(c = 0, p = source; c < len && *p != L'\0'; c++, p++) {
		dest[c] = *p;
	}

	if(c < len && *p == L'\0') {
		dest[c] = L'\0';
		return c;
	}

	while(*p) { c++; p++; }
	return c + 1;
}

VOID ClearDllDirectory() {
	PWSTR buffer, path, remove_p, in_p, out_p;
	RtlEnterCriticalSection(&dllDirectoryLock);
	if(dllDirectoryW[0] == L'\0') {
		RtlLeaveCriticalSection(&dllDirectoryLock);
		return;
	}
	
	buffer = (PWSTR)HeapAlloc(GetProcessHeap(), 0, sizeof(WCHAR) * MAX_ENV * 2);
	path   = buffer + MAX_ENV;
	
	GetEnvironmentVariableW(L"PATH", path, MAX_ENV);
	remove_p = wcsstr(path, dllDirectoryW);
	if(remove_p) {
		in_p = path; out_p = buffer;
		while(in_p != remove_p) { *out_p = *in_p; out_p++; in_p++; }
		in_p = dllDirectoryW;
		out_p--; //Remove the ; which we always add
		while(*in_p) { in_p++; remove_p++; }
		while(*remove_p == L';') { remove_p++; }
		in_p = remove_p;
		while(*in_p) { *out_p = *in_p; out_p++; in_p++; }
		*out_p = L'\0';
		SetEnvironmentVariableW(L"PATH", buffer);
	}
	
	dllDirectoryW[0] = L'\0';
	HeapFree(GetProcessHeap(), 0, buffer);
	RtlLeaveCriticalSection(&dllDirectoryLock);
}

DWORD WINAPI XP_GetDllDirectoryA(DWORD nBufferLength, LPSTR lpBuffer) {
	CHAR buffer[MAX_PATH];
	
	RtlEnterCriticalSection(&dllDirectoryLock);
	WideCharToMultiByte(CP_ACP, 0, dllDirectoryW, -1, buffer, ARRAY_SIZE(buffer), NULL, NULL);
	RtlLeaveCriticalSection(&dllDirectoryLock);

	return GetStringA(nBufferLength, lpBuffer, buffer);
}

DWORD WINAPI XP_GetDllDirectoryW(DWORD nBufferLength, LPWSTR lpBuffer) {
	DWORD dw;

	RtlEnterCriticalSection(&dllDirectoryLock);
	dw = GetStringW(nBufferLength, lpBuffer, dllDirectoryW);
	RtlLeaveCriticalSection(&dllDirectoryLock);

	return dw;
}

BOOL WINAPI XP_SetDllDirectoryW(LPCWSTR lpPathName) {
	PWSTR buffer, path;
	DWORD dw;
	
	if(!lpPathName) {
		ClearDllDirectory();
		return TRUE;
	}
	
	if(lstrcmpW(lpPathName, L"") == 0) {
		DbgPrintf(DBG_WARN, S_YELLOW "kernel32: Was supposed to remove the current directory from PATH\n");
		return TRUE;
	}
	
	ClearDllDirectory();
	
	buffer = (PWSTR)HeapAlloc(GetProcessHeap(), 0, sizeof(WCHAR) * MAX_ENV * 2);
	path   = buffer + MAX_ENV;
	
	RtlEnterCriticalSection(&dllDirectoryLock);
	dw = GetStringW(ARRAY_SIZE(dllDirectoryW), dllDirectoryW, lpPathName);
	if(dw > ARRAY_SIZE(dllDirectoryW)) {
		dllDirectoryW[0] = L'\0';
		HeapFree(GetProcessHeap(), 0, buffer);
		RtlLeaveCriticalSection(&dllDirectoryLock);
		return FALSE;
	}
	RtlLeaveCriticalSection(&dllDirectoryLock);
	
	GetEnvironmentVariableW(L"PATH", buffer, MAX_ENV);
	if(snwprintf(path, MAX_ENV, L"%s;%s", buffer, lpPathName) == -1)  {
		HeapFree(GetProcessHeap(), 0, buffer);
		return FALSE;
	}
	SetEnvironmentVariableW(L"PATH", path);

	HeapFree(GetProcessHeap(), 0, buffer);
	return TRUE;
}

BOOL WINAPI XP_SetDllDirectoryA(LPSTR lpPathName) {
	WCHAR buffer[MAX_PATH];
	
	if(!lpPathName) {
		ClearDllDirectory();
		return TRUE;
	}
	
	if(MultiByteToWideChar(CP_ACP, 0, lpPathName, -1, buffer, ARRAY_SIZE(buffer)) == 0) {
		return FALSE;
	}
	return XP_SetDllDirectoryW(buffer);
}

/*
77F85BD7  |. 64:A1 18000000 MOV EAX,DWORD PTR FS:[18]
77F85BDD  |. 8985 A4FDFFFF  MOV DWORD PTR SS:[EBP-25C],EAX
77F85BE3  |. 8B40 30        MOV EAX,DWORD PTR DS:[EAX+30]
77F85BE6  |. FFB0 A0000000  PUSH DWORD PTR DS:[EAX+A0]
77F85BEC  |. E8 EF220000    CALL ntdll.RtlEnterCriticalSection
*/

//DANGER WILL ROBINSON DANGER!!!!!!!!
//BESURE NOT TO CALL ANYTHING ELSE THAT COULD USE THE LOADER CODE WHILE
//LOCKED!
VOID LoaderLock(BOOL lock) {
	PROCESS_BASIC_INFORMATION info;
	NTSTATUS status;
	
	status = NtQueryInformationProcess(
	    GetCurrentProcess(), ProcessBasicInformation, &info, sizeof(info), NULL);
		
	if(!NT_SUCCESS(status)) return;
	
	if(lock)
		RtlEnterCriticalSection(*(PRTL_CRITICAL_SECTION *)((PBYTE)info.PebBaseAddress + 0xA0));
	else
		RtlLeaveCriticalSection(*(PRTL_CRITICAL_SECTION *)((PBYTE)info.PebBaseAddress + 0xA0));
}

BOOL IncLoadCount(HMODULE hMod) {
	WCHAR buffer[MAX_PATH+1];
	DWORD nSize;
	nSize = GetModuleFileNameW(hMod, buffer, MAX_PATH+1);
	if(nSize <= MAX_PATH) {
		if(LoadLibraryW(buffer)) return TRUE;
		else                     return FALSE;
	}
	return FALSE;
}

PLDR_MODULE WINAPI GetLdrModule(HMODULE hMod) {
	PROCESS_BASIC_INFORMATION info;
	PLDR_MODULE mod;
	PLDR_MODULE first_mod = NULL;
	
	NtQueryInformationProcess(
	    GetCurrentProcess(), ProcessBasicInformation, &info, sizeof(info), NULL);

	mod = (PLDR_MODULE)((PPEB_)info.PebBaseAddress)->LoaderData->InLoadOrderModuleList.Flink;
	while(mod != first_mod) {
		if(mod == first_mod) break;
		if(mod->BaseAddress == hMod) return mod;
		if(!first_mod) first_mod = mod;
		mod = (PLDR_MODULE)mod->InLoadOrderModuleList.Flink;
	}

	return NULL;
}

HMODULE WrapGetModule(HMODULE hMod) {
	if(GetModuleOriginal && hMod == thisModule) return hookedModule;
	return hMod;
}

HMODULE WINAPI XP_GetModuleHandleW(PWSTR lpModuleName) {
	return WrapGetModule(GetModuleHandleW(lpModuleName));
}

HMODULE WINAPI XP_GetModuleHandleA(LPCSTR lpModuleName) {
	return WrapGetModule(GetModuleHandleA(lpModuleName));
}

BOOL WINAPI XP_GetModuleHandleExW(DWORD dwFlags, LPCWSTR lpModuleName, HMODULE* phModule) {
	PROCESS_BASIC_INFORMATION info;
	PLDR_MODULE mod;
	PLDR_MODULE first_mod = NULL;
	BOOL inc_loadcount = FALSE;

	*phModule = NULL;
	
	LoaderLock(TRUE);
	if(dwFlags & GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS) {
		NtQueryInformationProcess(
		    GetCurrentProcess(), ProcessBasicInformation, &info, sizeof(info), NULL);
		mod = (PLDR_MODULE)((PPEB_)info.PebBaseAddress)->LoaderData->InLoadOrderModuleList.Flink;
		while(mod != first_mod) {
			if((DWORD)mod->BaseAddress <= (DWORD)lpModuleName && 
			   (DWORD)lpModuleName < (DWORD)mod->BaseAddress + (DWORD)mod->SizeOfImage) {
				*phModule = mod->BaseAddress;
				break;
			}
			if(!first_mod) first_mod = mod;
			mod = (PLDR_MODULE)mod->InLoadOrderModuleList.Flink;
		}
	} else {
		*phModule = GetModuleHandleW(lpModuleName);
	}
	LoaderLock(FALSE);

	if(*phModule == NULL) return FALSE;
	*phModule = WrapGetModule(*phModule);

	LoaderLock(TRUE);
	mod = GetLdrModule(*phModule);
	if(!(dwFlags & GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT))
		inc_loadcount = TRUE;

	//How to pin? We'll just inc the LoadCount and hope
	if(dwFlags & GET_MODULE_HANDLE_EX_FLAG_PIN) 
		inc_loadcount = TRUE;
	LoaderLock(FALSE);
	
	if(inc_loadcount) {
		IncLoadCount(*phModule);
	}
	
	return TRUE;
}

BOOL WINAPI XP_GetModuleHandleExA(DWORD dwFlags, LPCSTR lpModuleName, HMODULE* phModule) {
	BOOL ret;
	UNICODE_STRING unicode;
	STRING ansi;
	if(!(dwFlags & GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS) && lpModuleName) {
		RtlInitString(&ansi, lpModuleName);
		RtlAnsiStringToUnicodeString(&unicode, &ansi, TRUE);
		ret = XP_GetModuleHandleExW(dwFlags, unicode.Buffer, phModule);
		RtlFreeUnicodeString(&unicode);
	} else {
		ret = XP_GetModuleHandleExW(dwFlags, (LPCWSTR)lpModuleName, phModule);
	}	
	return ret;
}

BOOL WINAPI ShouldApplyVersionFaking(PVOID address) {
	HMODULE hMod;
	int i;
	if(XP_GetModuleHandleExW(
		GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS | GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT,
		(LPCWSTR)address, &hMod)) {
		for(i = 0; i < numFilterModules; i++) {
			if(!filterModules[i]) filterModules[i] = XP_GetModuleHandleA(filterModuleNames[i]);
			if(hMod == filterModules[i]) return versionFaking == VERSION_FAKING_INCLUDE;
		}
	}
	return versionFaking == VERSION_FAKING_EXCLUDE;
}

DWORD WINAPI XP_GetProcessId(HANDLE Process) {
	PROCESS_BASIC_INFORMATION info;
	ULONG len;
	NTSTATUS status;
	
	status = NtQueryInformationProcess(
	    Process, ProcessBasicInformation, &info, sizeof(info), &len);

	if(NT_SUCCESS(status)) {
		return info.UniqueProcessId;
	}
	
	SetLastError(RtlNtStatusToDosError(status));
	return 0;
}

BOOL WINAPI XP_GetProcessHandleCount(HANDLE hProcess, PDWORD pdwHandleCount) {
	ULONG len;
	NTSTATUS status;
	
	status = NtQueryInformationProcess(
	    hProcess, ProcessHandleCount, pdwHandleCount, sizeof(DWORD), &len);
		
	if(NT_SUCCESS(status)) {
		return TRUE;
	}
	
	SetLastError(RtlNtStatusToDosError(status));
	return FALSE;
}

DWORD WINAPI XP_GetThreadId(HANDLE Thread) {
    THREAD_BASIC_INFORMATION tbi;
    ULONG len;
    NTSTATUS status;

    status = NtQueryInformationThread(
		Thread, ThreadBasicInformation, &tbi, sizeof(tbi), &len);

    if (NT_SUCCESS(status)) {
		return (DWORD)tbi.ClientId.UniqueThread;
    }

    SetLastError(RtlNtStatusToDosError(status));
    return 0;
}

void WINAPI XP_RestoreLastError(DWORD dwErrCode) {
	SetLastError(dwErrCode);	
}

/*
BOOL WINAPI XP_HeapSetInformation(
	HANDLE HeapHandle,
	HEAP_INFORMATION_CLASS HeapInformationClass,
	PVOID HeapInformation,
	SIZE_T HeapInformationLength) {
	//This function only supports HeapCompatibilityInformation on XP and
	//only is used to enable LFH.
	if(HeapInformationLength == sizeof(ULONG) && *(ULONG *)HeapInformation == 2) return TRUE;
	return FALSE;
}
*/

DWORD WINAPI XP_GetVersion(void) {
	PVOID callee_address;
	callee_address = _ReturnAddress();

	if(ShouldApplyVersionFaking(callee_address))
		return versionInfoA.dwMajorVersion | (versionInfoA.dwMinorVersion << 8) | (versionInfoA.dwBuildNumber << 16);
	else return GetVersion();
}

BOOL WINAPI XP_GetVersionExA(LPOSVERSIONINFOA lpVersionInfo) {
	PVOID callee_address;
	callee_address = _ReturnAddress();
	
	DbgPrintf(DBG_TRACE, "kernel32: GetVersionExA from %p\n", callee_address);
	
	if(ShouldApplyVersionFaking(callee_address) &&
	   (lpVersionInfo->dwOSVersionInfoSize == sizeof(OSVERSIONINFOA) ||
		lpVersionInfo->dwOSVersionInfoSize == sizeof(OSVERSIONINFOEXA))) {
		memcpy(lpVersionInfo, &versionInfoA, lpVersionInfo->dwOSVersionInfoSize);
		return TRUE;
	}
	
	return GetVersionExA(lpVersionInfo);
}

BOOL WINAPI XP_GetVersionExW(LPOSVERSIONINFOW lpVersionInfo) {
	PVOID callee_address;
	callee_address =_ReturnAddress();
	
	DbgPrintf(DBG_TRACE, "kernel32: GetVersionExW from %p\n", callee_address);
	
	if(ShouldApplyVersionFaking(callee_address) &&
	   (lpVersionInfo->dwOSVersionInfoSize == sizeof(OSVERSIONINFOW) ||
		lpVersionInfo->dwOSVersionInfoSize == sizeof(OSVERSIONINFOEXW))) {
		memcpy(lpVersionInfo, &versionInfoW, lpVersionInfo->dwOSVersionInfoSize);
		return TRUE;
	}
	
	return GetVersionExW(lpVersionInfo);
}


BOOL WINAPI VerCondChk(DWORD cond, DWORD sys_val, DWORD chk_val) {
	switch(cond) {
		case VER_EQUAL:         return (sys_val == chk_val) ? TRUE : FALSE;
		case VER_GREATER:       return (sys_val >  chk_val) ? TRUE : FALSE;
		case VER_GREATER_EQUAL: return (sys_val >= chk_val) ? TRUE : FALSE;
		case VER_LESS:          return (sys_val <  chk_val) ? TRUE : FALSE;
		case VER_LESS_EQUAL:    return (sys_val <= chk_val) ? TRUE : FALSE;
		case VER_AND:           return (sys_val &  chk_val) == chk_val ? TRUE : FALSE;
		case VER_OR:            return (sys_val &  chk_val) ? TRUE : FALSE;
		default:                return FALSE;
	}
}

BOOL WINAPI XP_VerifyVersionInfoA(LPOSVERSIONINFOEXA lpVersionInfo, DWORD dwTypeMask, DWORDLONG dwlConditionMask) {
	DWORD cond, type, i;
	BOOL result;
	PVOID callee_address;
	callee_address = _ReturnAddress();
	
	DbgPrintf(DBG_TRACE, "kernel32: VerifyVersionInfoA from %p\n", callee_address);

	if(ShouldApplyVersionFaking(callee_address)) {
		if(verifyVersion) return TRUE;
		result = TRUE;
		for(i = 0; i < 8; i++) {
			type = 1 << i;
			if(!(dwTypeMask & type)) continue;
			cond = (DWORD)(dwlConditionMask >> (i*VER_NUM_BITS_PER_CONDITION_MASK)) & VER_CONDITION_MASK;
			switch(type) {
				case VER_MINORVERSION:     result &= VerCondChk(cond, versionInfoA.dwMinorVersion, lpVersionInfo->dwMinorVersion); break;
				case VER_MAJORVERSION:     result &= VerCondChk(cond, versionInfoA.dwMajorVersion, lpVersionInfo->dwMajorVersion); break;
				case VER_BUILDNUMBER:      result &= VerCondChk(cond, versionInfoA.dwBuildNumber, lpVersionInfo->dwBuildNumber); break;
				case VER_PLATFORMID:       result &= VerCondChk(cond, versionInfoA.dwPlatformId, lpVersionInfo->dwPlatformId); break;
				case VER_SERVICEPACKMAJOR: result &= VerCondChk(cond, versionInfoA.wServicePackMajor, lpVersionInfo->wServicePackMajor); break;
				case VER_SERVICEPACKMINOR: result &= VerCondChk(cond, versionInfoA.wServicePackMinor, lpVersionInfo->wServicePackMinor); break;
				case VER_SUITENAME:        result &= VerCondChk(cond, versionInfoA.wSuiteMask, lpVersionInfo->wSuiteMask); break;
				case VER_PRODUCT_TYPE:     result &= VerCondChk(cond, versionInfoA.wProductType, lpVersionInfo->wProductType); break;
				default: continue;
			}
			if(!result) {
				SetLastError(ERROR_OLD_WIN_VERSION);
				return FALSE;
			}
		}
		return result;
	}	
	return VerifyVersionInfoA(lpVersionInfo, dwTypeMask, dwlConditionMask);
}

BOOL WINAPI XP_VerifyVersionInfoW(LPOSVERSIONINFOEXW lpVersionInfo, DWORD dwTypeMask, DWORDLONG dwlConditionMask) {
	DWORD cond, type, i;
	BOOL result;
	PVOID callee_address;
	callee_address = _ReturnAddress();
	
	DbgPrintf(DBG_TRACE, "kernel32: VerifyVersionInfoW from %p\n", callee_address);

	if(ShouldApplyVersionFaking(callee_address)) {
		if(verifyVersion) return TRUE;
		result = TRUE;
		for(i = 0; i < 8; i++) {
			type = 1 << i;
			if(!(dwTypeMask & type)) continue;
			cond = (DWORD)(dwlConditionMask >> (i*VER_NUM_BITS_PER_CONDITION_MASK)) & VER_CONDITION_MASK;
			switch(type) {
				case VER_MINORVERSION:     result &= VerCondChk(cond, versionInfoA.dwMinorVersion, lpVersionInfo->dwMinorVersion); break;
				case VER_MAJORVERSION:     result &= VerCondChk(cond, versionInfoA.dwMajorVersion, lpVersionInfo->dwMajorVersion); break;
				case VER_BUILDNUMBER:      result &= VerCondChk(cond, versionInfoA.dwBuildNumber, lpVersionInfo->dwBuildNumber); break;
				case VER_PLATFORMID:       result &= VerCondChk(cond, versionInfoA.dwPlatformId, lpVersionInfo->dwPlatformId); break;
				case VER_SERVICEPACKMAJOR: result &= VerCondChk(cond, versionInfoA.wServicePackMajor, lpVersionInfo->wServicePackMajor); break;
				case VER_SERVICEPACKMINOR: result &= VerCondChk(cond, versionInfoA.wServicePackMinor, lpVersionInfo->wServicePackMinor); break;
				case VER_SUITENAME:        result &= VerCondChk(cond, versionInfoA.wSuiteMask, lpVersionInfo->wSuiteMask); break;
				case VER_PRODUCT_TYPE:     result &= VerCondChk(cond, versionInfoA.wProductType, lpVersionInfo->wProductType); break;
				default: continue;
			}
			if(!result) {
				SetLastError(ERROR_OLD_WIN_VERSION);
				return FALSE;
			}
		}
		return result;
	}	
	return VerifyVersionInfoW(lpVersionInfo, dwTypeMask, dwlConditionMask);
}

PVOID WINAPI XP_EncodePointer(PVOID ptr) {
	return (PVOID)((UINT_PTR)ptr ^ 0xDEADBEEF);
}

PVOID WINAPI XP_DecodePointer(PVOID ptr) {
	return (PVOID)((UINT_PTR)ptr ^ 0xDEADBEEF);
}

BYTE slist_lock[0x100];
void SListLock(PSLIST_HEADER ListHead) {
	DWORD index = (((DWORD)ListHead) >> MEMORY_ALLOCATION_ALIGNMENT) & 0xFF;
	
	__asm {
		mov edx, index
		mov cl, 1
  spin:	mov al, 0
		lock cmpxchg byte ptr [slist_lock + edx], cl
		jnz spin
	};

	return;
}

void SListUnlock(PSLIST_HEADER ListHead) {
	DWORD index = (((DWORD)ListHead) >> MEMORY_ALLOCATION_ALIGNMENT) & 0xFF;
	slist_lock[index] = 0;
}

USHORT WINAPI XP_QueryDepthSList(PSLIST_HEADER ListHead) {
	PSLIST_ENTRY n;
	USHORT depth = 0;
	SListLock(ListHead);
	n = ListHead->Next.Next;
	while(n) {
		depth++;
		n = n->Next;
	}
	SListUnlock(ListHead);
	return depth;
}

//Note: ListHead->Next.Next == first node
PSLIST_ENTRY WINAPI XP_InterlockedPopEntrySList(PSLIST_HEADER ListHead) {
	PSLIST_ENTRY ret;
	SListLock(ListHead);
	if(!ListHead->Next.Next) ret = NULL;
	else { 
		ret = ListHead->Next.Next;
		ListHead->Next.Next = ret->Next;
		ListHead->Depth--;
		ListHead->Sequence++;
	}
	SListUnlock(ListHead);
	return ret;
}

PSLIST_ENTRY WINAPI XP_InterlockedPushEntrySList(PSLIST_HEADER ListHead, PSLIST_ENTRY ListEntry) {
	PSLIST_ENTRY ret;
	SListLock(ListHead);
	ret = ListHead->Next.Next;
	ListEntry->Next = ret;
	ListHead->Next.Next = ListEntry;
	ListHead->Depth++;
	ListHead->Sequence++;
	SListUnlock(ListHead);
	return ret;
}

PSLIST_ENTRY WINAPI XP_InterlockedFlushSList(PSLIST_HEADER ListHead) {
	PSLIST_ENTRY ret;
	SListLock(ListHead);
	ret = ListHead->Next.Next;
	ListHead->Next.Next = NULL;
	ListHead->Depth = 0;
	ListHead->Sequence++;
	SListUnlock(ListHead);
	return ret;
}

void WINAPI XP_InitializeSListHead(PSLIST_HEADER ListHead) {
	SListLock(ListHead);
	ListHead->Next.Next = NULL;
	ListHead->Depth = 0;
	ListHead->Sequence = 0;
	SListUnlock(ListHead);
}

void WINAPI XP_GetNativeSystemInfo(LPSYSTEM_INFO lpSystemInfo) {
	GetSystemInfo(lpSystemInfo);
	return;
}

DWORD WINAPI XP_WTSGetActiveConsoleSessionId(void) {
	return 0;
}

GEOID WINAPI XP_GetUserGeoID(GEOCLASS GeoClass) {
	switch(GeoClass) {
		case GEOCLASS_NATION: return geoNation;
		case GEOCLASS_REGION: return geoRegion;
		default: 
			DbgPrintf(DBG_WARN, S_YELLOW "kernel32: GetUserGeoID unimplemented GEOCLASS %X\n", GeoClass);
			return GEOID_NOT_AVAILABLE;
	}
}

/* 47414b !   call        dword ptr [KERNEL32.dll:GetUserGeoID]
 474151 !   mov         edi, eax
 474153 !   cmp         edi, 0ffffffffh
 474156 !   jz          loc_4741d5
 474158 !   mov         esi, [KERNEL32.dll:GetGeoInfoA]
 47415e !   xor         eax, eax
 474160 !   push        eax ;LangId
 474161 !   push        eax ;cchData
 474162 !   push        eax ;lpGeoDate
 474163 !   push        4 ;GeoType=GEO_ISO2
 474165 !   push        edi ;Location*/
 
int WINAPI XP_GetGeoInfoW(GEOID Location, GEOTYPE GeoType, LPWSTR lpGeoData, int cchData, LANGID LangId) {
	int size;
	
	if(LangId == 0 && (GeoType == GEO_LCID || GeoType == GEO_RFC1766)) {
		LangId = GetUserDefaultLangID();
	}
	
	switch(GeoType) {
		case GEO_NATION:
			if(cchData == 0) return sizeof(long) / sizeof(WORD);
			if(cchData < sizeof(long) / sizeof(WORD)) {
				SetLastError(ERROR_INSUFFICIENT_BUFFER);
				return 0;
			}
			*((long *)lpGeoData) = Location;
			return sizeof(long) / sizeof(WORD);
		case GEO_LATITUDE:
		case GEO_LONGITUDE:
			if(cchData == 0) return sizeof(float) / sizeof(WORD);
			if(cchData < sizeof(float) / sizeof(WORD)) {
				SetLastError(ERROR_INSUFFICIENT_BUFFER);
				return 0;
			}
			*((float *)lpGeoData) = 0.0f;
			return sizeof(float) / sizeof(WORD);
		case GEO_ISO2:
			if(cchData == 0) return ARRAY_SIZE(geoISO2W);
			if(cchData < ARRAY_SIZE(geoISO2W)) {
				SetLastError(ERROR_INSUFFICIENT_BUFFER);
				return 0;
			}
			memcpy(lpGeoData, geoISO2W, sizeof(geoISO2W));
			return ARRAY_SIZE(geoISO2W);
		case GEO_ISO3:
			if(cchData == 0) return ARRAY_SIZE(geoISO3W);
			if(cchData < ARRAY_SIZE(geoISO3W)) {
				SetLastError(ERROR_INSUFFICIENT_BUFFER);
				return 0;
			}
			memcpy(lpGeoData, geoISO3W, sizeof(geoISO3W));
			return ARRAY_SIZE(geoISO3W);
		case GEO_RFC1766:
			size = lstrlenW(geoRFC1766W) + 1;
			if(cchData == 0) return size;
			if(cchData < size) {
				SetLastError(ERROR_INSUFFICIENT_BUFFER);
				return 0;
			}
			memcpy(lpGeoData, geoRFC1766W, size * sizeof(WCHAR));
			return size;
		case GEO_LCID:
			if(cchData == 0) return sizeof(DWORD) / sizeof(WORD);
			if(cchData < sizeof(DWORD) / sizeof(WORD)) {
				SetLastError(ERROR_INSUFFICIENT_BUFFER);
				return 0;
			}
			*((DWORD *)lpGeoData) = MAKELCID(LangId, SORT_DEFAULT);
			return sizeof(DWORD) / sizeof(WORD);
		case GEO_FRIENDLYNAME:
			size = lstrlenW(geoFriendlyW) + 1;
			if(cchData == 0) return size;
			if(cchData < size) {
				SetLastError(ERROR_INSUFFICIENT_BUFFER);
				return 0;
			}
			memcpy(lpGeoData, geoFriendlyW, size * sizeof(WCHAR));
			return size;
		case GEO_OFFICIALNAME:
			size = lstrlenW(geoOfficialW) + 1;
			if(cchData == 0) return size;
			if(cchData < size) {
				SetLastError(ERROR_INSUFFICIENT_BUFFER);
				return 0;
			}
			memcpy(lpGeoData, geoOfficialW, size * sizeof(WCHAR));
			return size;
		default:
			SetLastError(ERROR_INVALID_PARAMETER);
			return 0;
	}
}

int WINAPI XP_GetGeoInfoA(GEOID Location, GEOTYPE GeoType, LPSTR lpGeoData, int cchData, LANGID LangId) {
	WCHAR buffer[128];
	int r;
	
	switch(GeoType) {
		case GEO_ISO2:
		case GEO_ISO3:
		case GEO_RFC1766:
		case GEO_FRIENDLYNAME:
		case GEO_OFFICIALNAME:
			r = XP_GetGeoInfoW(Location, GeoType, NULL, 0, LangId);
			if(r == 0 || cchData == 0) return r;
			if(cchData < r) {
				SetLastError(ERROR_INSUFFICIENT_BUFFER);
				return 0;
			}
			r = XP_GetGeoInfoW(Location, GeoType, buffer, ARRAY_SIZE(buffer), LangId);
			if(r) return WideCharToMultiByte(CP_ACP, 0, buffer, -1, lpGeoData, cchData, NULL, NULL);
			else  return r;
		
		case GEO_NATION:
		case GEO_LATITUDE:
		case GEO_LONGITUDE:
		case GEO_LCID:
			return XP_GetGeoInfoW(Location, GeoType, (LPWSTR)lpGeoData, cchData / sizeof(WCHAR), LangId);
		
		default: 
			SetLastError(ERROR_INVALID_PARAMETER);
			return 0;
	}
}


UINT WINAPI XP_GetSystemDirectoryW(PWSTR lpBuffer, UINT uSize) {
	UINT i;
	for(i = 0; systemDirectoryW[i]; i++) {
		if(i < uSize) lpBuffer[i] = systemDirectoryW[i];
	}
	if(i < uSize) { lpBuffer[i] = L'\0'; return i; }
	return i + 1;
}

UINT WINAPI XP_GetSystemDirectoryA(LPSTR lpBuffer, UINT uSize) {
	UINT i;
	for(i = 0; systemDirectoryA[i]; i++) {
		if(i < uSize) lpBuffer[i] = systemDirectoryA[i];
	}
	if(i < uSize) { lpBuffer[i] = '\0'; return i; }
	return i + 1;
}

BOOL WINAPI XP_GetNumaNodeProcessorMask(UCHAR Node, PULONGLONG ProcessorMask) {
	DWORD_PTR ProcessAffinity;
	DWORD_PTR SystemAffinity;
	
	if(GetProcessAffinityMask(GetCurrentProcess(), &ProcessAffinity, &SystemAffinity)) {
		*ProcessorMask = SystemAffinity;
		return TRUE;
	} else {
		return FALSE;
	}
}

BOOL WINAPI XP_GetNumaProcessorNode(UCHAR Processor, PUCHAR NodeNumber) {
	DWORD_PTR ProcessAffinity;
	DWORD_PTR SystemAffinity;
	
	if(GetProcessAffinityMask(GetCurrentProcess(), &ProcessAffinity, &SystemAffinity)) {
		if(1<<Processor & SystemAffinity) *NodeNumber = 0;
		else *NodeNumber = 0xFF;
		return TRUE;
	} else {
		return FALSE;
	}
}

BOOL WINAPI XP_GetVolumePathNamesForVolumeNameA(
  LPCSTR lpszVolumeName, LPSTR lpszVolumePathNames, DWORD cchBufferLength, PDWORD lpcchReturnLength) {
	CHAR buffer[MAX_PATH+16];
	HANDLE findVMP;
	BOOL roomLeft = TRUE;
	int len;
	
	*lpcchReturnLength = 0;
	
	findVMP = FindFirstVolumeMountPointA(lpszVolumeName, buffer, sizeof(buffer) / sizeof(CHAR));
	if(findVMP == INVALID_HANDLE_VALUE) {
		return FALSE;
	}
	
	len = lstrlenA(buffer);
	if(*lpcchReturnLength + len + 2 < cchBufferLength) {
		lstrcpyA(lpszVolumePathNames, buffer);
		lpszVolumePathNames += (len + 1);
	} else {
		roomLeft = FALSE;
	}
	*lpcchReturnLength += (len + 1);
			
	while(FindNextVolumeMountPointA(findVMP, buffer, sizeof(buffer) / sizeof(CHAR))) {
		len = lstrlenA(buffer);
		if(*lpcchReturnLength + len + 2 < cchBufferLength) {
			lstrcpyA(lpszVolumePathNames, buffer);
			lpszVolumePathNames += (len + 1);
		} else {
			roomLeft = FALSE;;
		}
		*lpcchReturnLength += (len + 1);
	}
	*lpcchReturnLength += 1;
	
	if(GetLastError() == ERROR_NO_MORE_FILES && roomLeft) {
		*lpszVolumePathNames = '\0';
	}
	
	FindVolumeMountPointClose(findVMP);
	
	if(!roomLeft) SetLastError(ERROR_MORE_DATA);
	return roomLeft;
}

BOOL WINAPI XP_GetVolumePathNamesForVolumeNameW(
	LPCWSTR lpszVolumeName, LPWSTR lpszVolumePathNames, DWORD cchBufferLength, PDWORD lpcchReturnLength) {
	WCHAR buffer[MAX_PATH+16];
	HANDLE findVMP;
	BOOL roomLeft = TRUE;
	int len;
	
	*lpcchReturnLength = 0;
	
	findVMP = FindFirstVolumeMountPointW(lpszVolumeName, buffer, sizeof(buffer) / sizeof(WCHAR));
	if(findVMP == INVALID_HANDLE_VALUE) {
		return FALSE;
	}
	
	len = lstrlenW(buffer);
	if(*lpcchReturnLength + len + 2 < cchBufferLength) {
		lstrcpyW(lpszVolumePathNames, buffer);
		lpszVolumePathNames += (len + 1);
	} else {
		roomLeft = FALSE;
	}
	*lpcchReturnLength += (len + 1);
			
	while(FindNextVolumeMountPointW(findVMP, buffer, sizeof(buffer) / sizeof(WCHAR))) {
		len = lstrlenW(buffer);
		if(*lpcchReturnLength + len + 2 < cchBufferLength) {
			lstrcpyW(lpszVolumePathNames, buffer);
			lpszVolumePathNames += (len + 1);
		} else {
			roomLeft = FALSE;;
		}
		*lpcchReturnLength += (len + 1);
	}
	*lpcchReturnLength += 1;
	
	if(GetLastError() == ERROR_NO_MORE_FILES && roomLeft) {
		*lpszVolumePathNames = L'\0';
	}
	
	FindVolumeMountPointClose(findVMP);
	
	if(!roomLeft) SetLastError(ERROR_MORE_DATA);
	return roomLeft;
}

HMODULE WINAPI XP_LoadLibraryA(LPCSTR lpFileName) {
	HMODULE hmod = LoadLibraryA(lpFileName);
	return hmod;
}

HMODULE WINAPI XP_LoadLibraryW(LPCWSTR lpFileName) {
	HMODULE hmod = LoadLibraryW(lpFileName);
	return hmod;
}

HMODULE (WINAPI *_LoadLibraryExW)(LPCWSTR lpFileName, HANDLE hFile, DWORD dwFlags);
HMODULE WINAPI XP_LoadLibraryExW(LPCWSTR lpFileName, HANDLE hFile, DWORD dwFlags) {
	static const WCHAR *filter_dlls[] = {
		L"kernel32", L"user32", L"advapi32", L"shell32", L"ws2_32", 
		L"wtsapi32", L"imm32", L"uxtheme", L"imagehlp", L"hid"
	};
	const WCHAR *p, *q, *dll;
	int i;
	
	if(!lpFileName) return _LoadLibraryExW(lpFileName, hFile, dwFlags);
	for(p = dll = lpFileName; *p; p++) if(*p == L'\\' || *p == L'/') dll = p + 1;
	
	for(i=0; i<ARRAY_SIZE(filter_dlls); i++) {
		for(p = filter_dlls[i], q = dll;
		  *p; p++, q++) {
			if(!*q || towupper(*p) != towupper(*q)) goto nexti;
		}
		if(*q && _wcsicmp(L".dll", q) != 0) goto nexti;
		DbgPrintf(DBG_INFO, "kernel32: intercepted LoadLibrary(%S) redirecting to %S\n",
			lpFileName, filter_dlls[i]);
		return _LoadLibraryExW(filter_dlls[i], hFile, dwFlags);
		nexti:;
	}
	
	return _LoadLibraryExW(lpFileName, hFile, dwFlags);
}

FARPROC WINAPI XP_GetProcAddress(HMODULE hModule, LPCSTR lpProcName) {
	FARPROC proc;
#if 1
	WCHAR file_path[MAX_PATH];
	WCHAR *p, *file_name;
	HMODULE hMod;
	PVOID ra;
	
	if((((DWORD)lpProcName) >> 16) && (strcmp(lpProcName, "FindActCtxSectionStringW") == 0 || strcmp(lpProcName, "FindActCtxSectionStringA") == 0)) {
		ra = _ReturnAddress();
		if(XP_GetModuleHandleExW(GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS, ra, &hMod)) {
			if(GetModuleFileNameW(hMod, file_path, sizeof(file_path)/sizeof(WCHAR))) {
				for(file_name = p = file_path; *p; ++p) {
					if(*p == L'/' || *p == L'\\') file_name = p + 1;
				}
				if(*file_name) {
					if(_wcsnicmp(file_name, L"msvcr", 5) == 0) {
						DbgPrintf(DBG_INFO, "kernel32: GetProcAddress returning NULL to msvcr for FindActCtxSectionStringW\n");
						return NULL;
					}
				}
			}
		}
	}
#endif	
	proc = GetProcAddress(hModule, lpProcName);
	return proc;
}

BOOL WINAPI XP_CheckRemoteDebuggerPresent(HANDLE hProcess, PBOOL pbDebuggerPresent) {
	*pbDebuggerPresent = FALSE;
	return FALSE;
}

#define EVALRVA(type, x) ((type)((DWORD)hModule + (DWORD)(x)))
PIMAGE_DATA_DIRECTORY GetDataDirectory(HMODULE hModule, int directory) {
	PIMAGE_DOS_HEADER dosheader;
	PIMAGE_NT_HEADERS ntheader;
	dosheader = EVALRVA(PIMAGE_DOS_HEADER, 0);
	ntheader  = EVALRVA(PIMAGE_NT_HEADERS, dosheader->e_lfanew);
	return &ntheader->OptionalHeader.DataDirectory[directory];
}

void RenameExportTableName(HMODULE hModule, const char *name) {
	PIMAGE_DATA_DIRECTORY exportdd;
	PIMAGE_EXPORT_DIRECTORY exportd;
	DWORD oldProtect;
	
	exportdd  = GetDataDirectory(hModule, IMAGE_DIRECTORY_ENTRY_EXPORT);
	exportd   = EVALRVA(PIMAGE_EXPORT_DIRECTORY, exportdd->VirtualAddress);
	
	VirtualProtect(EVALRVA(DWORD *, exportd->Name), sizeof(DWORD), PAGE_EXECUTE_READWRITE, &oldProtect);
	*EVALRVA(DWORD *, exportd->Name) = (DWORD)name - (DWORD)hModule; //RVA from module
	VirtualProtect(EVALRVA(DWORD *, exportd->Name), sizeof(DWORD), oldProtect, &oldProtect);
}

void ForwardExportsToFunctionExports(HMODULE hModule) {
	PIMAGE_DATA_DIRECTORY exportdd;
	PIMAGE_EXPORT_DIRECTORY exportd;
	DWORD *func_addresses;
	DWORD func_address;
	DWORD oldProtect;
	UINT i;
	
	exportdd  = GetDataDirectory(hModule, IMAGE_DIRECTORY_ENTRY_EXPORT);
	exportd   = EVALRVA(PIMAGE_EXPORT_DIRECTORY, exportdd->VirtualAddress);
	func_addresses = EVALRVA(DWORD *, exportd->AddressOfFunctions);
	
	VirtualProtect(func_addresses, sizeof(DWORD) * exportd->NumberOfFunctions, PAGE_EXECUTE_READWRITE, &oldProtect);
	for(i = 0; i < exportd->NumberOfFunctions; i++) {
		if(exportdd->VirtualAddress <= func_addresses[i] &&
		  func_addresses[i] < exportdd->VirtualAddress + exportdd->Size) {
			func_address = (DWORD)GetProcAddress(hModule, (LPCSTR)i);
			if(func_address)
				func_addresses[i - exportd->Base] = func_address - (DWORD)hModule; //RVA from module
		}
	}
	VirtualProtect(func_addresses, sizeof(DWORD) * exportd->NumberOfFunctions, oldProtect, &oldProtect);
}
#undef EVALRVA

/*
BOOL WINAPI XP_GetSystemTimes(LPFILETIME lpIdleTime, LPFILETIME lpKernelTime, LPFILETIME lpUserTime) {
	SYSTEM_PROCESSOR_PERFORMANCE_INFORMATION info;
	NTSTATUS status;
	ULONG len;
	status = NtQuerySystemInformation(
		SystemProcessorPerformanceInformation, &info, sizeof(info), &len);
	if(NT_SUCCESS(status)) {
		if(lpKernelTime) *lpKernelTime = *(PFILETIME)&info.KernelTime;
		if(lpUserTime)   *lpUserTime   = *(PFILETIME)&info.UserTime;
		if(lpIdleTime)   *lpIdleTime   = *(PFILETIME)&info.IdleTime;
		return TRUE;
	} else {
		SetLastError(RtlNtStatusToDosError(status));
		return FALSE;
	}
}*/

/* Multiple core CPU support from TheProphet*/
BOOL WINAPI XP_GetSystemTimes(LPFILETIME lpIdleTime, LPFILETIME lpKernelTime, LPFILETIME lpUserTime) {
    SYSTEM_BASIC_INFORMATION                    SysBasicInfo;
    SYSTEM_PROCESSOR_PERFORMANCE_INFORMATION  * SysProcPerfInfo = NULL;
    NTSTATUS status;
    LARGE_INTEGER it = {0, 0}, kt = {0, 0}, ut = {0, 0};
    int i;
    HANDLE hHeap = NULL;
    
	//Need atleast one out parameter
    if(!lpIdleTime && !lpKernelTime && !lpUserTime)
        return FALSE;
    
	//Get number of processors
    status = NtQuerySystemInformation(SystemBasicInformation, &SysBasicInfo, sizeof(SysBasicInfo), NULL);
    if(NT_SUCCESS(status))
    {
		hHeap = GetProcessHeap();
        SysProcPerfInfo = (SYSTEM_PROCESSOR_PERFORMANCE_INFORMATION*)
			RtlAllocateHeap(hHeap, 0,
				SysBasicInfo.NumberOfProcessors * sizeof(SYSTEM_PROCESSOR_PERFORMANCE_INFORMATION));
        
        if(!SysProcPerfInfo) return FALSE;
        
		//Get counters
        status = NtQuerySystemInformation(SystemProcessorPerformanceInformation, SysProcPerfInfo, 
			sizeof(SYSTEM_PROCESSOR_PERFORMANCE_INFORMATION)*SysBasicInfo.NumberOfProcessors, NULL);
			
        if(NT_SUCCESS(status))
        {
            for(i=0; i<SysBasicInfo.NumberOfProcessors; i++)
            {
                it.QuadPart += SysProcPerfInfo[i].IdleTime.QuadPart;
                kt.QuadPart += SysProcPerfInfo[i].KernelTime.QuadPart;
                ut.QuadPart += SysProcPerfInfo[i].UserTime.QuadPart;
            }
            
            if(lpIdleTime)
            {
                lpIdleTime->dwLowDateTime = it.LowPart;
                lpIdleTime->dwHighDateTime = it.HighPart;
            }
            
            if(lpKernelTime)
            {
                lpKernelTime->dwLowDateTime = kt.LowPart;
                lpKernelTime->dwHighDateTime = kt.HighPart;
            }

            if(lpUserTime)
            {
                lpUserTime->dwLowDateTime = ut.LowPart;
                lpUserTime->dwHighDateTime = ut.HighPart;
            }            
            
            RtlFreeHeap(hHeap, 0, SysProcPerfInfo);
            return TRUE;
        } else {
			RtlFreeHeap(hHeap, 0, SysProcPerfInfo);
		}
    }        
    
    SetLastError(RtlNtStatusToDosError(status));
    return FALSE;
}


UINT WINAPI XP_GetSystemWow64DirectoryA(LPSTR lpBuffer, UINT uSize) {
	SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
	return 0;
}

UINT WINAPI XP_GetSystemWow64DirectoryW(LPWSTR lpBuffer, UINT uSize) {
	SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
	return 0;
}

BOOL WINAPI XP_IsWow64Process(HANDLE hProcess, PBOOL Wow64Process) {
	*Wow64Process = FALSE;
	return TRUE;
}

BOOL WINAPI XP_AttachConsole(DWORD dwProcessId) {
	if (dwProcessId != GetCurrentProcessId())
		DbgPrintf(DBG_WARN, S_YELLOW "kernel32: AttachConsole does not support other processes");
	//Just make sure we have a console
	AllocConsole();
	return TRUE;
}

VOID  __declspec(naked) WINAPI XP_RtlCaptureContext(PCONTEXT ContextRecord) {
	__asm {
		push eax
		push edx
		pushfd
		
		mov  eax, [esp + 16]
		
		//DWORD Flags
        mov  [eax], CONTEXT_FULL
		add  eax, 4

		//6 DWORD debug registers
#if 0
		mov  edx, dr0
		mov  [eax], edx
		add  eax, 4
		
		mov  edx, dr1
		mov  [eax], edx
		add  eax, 4
		
		mov  edx, dr2
		mov  [eax], edx
		add  eax, 4
		
		mov  edx, dr3
		mov  [eax], edx
		add  eax, 4
		
		mov  edx, dr6
		mov  [eax], edx
		add  eax, 4
		
		mov  edx, dr7
		mov  [eax], edx
		add  eax, 4
#else		
		add  eax, 4*6
#endif
		
		//Floating Point
		fsave  [eax]
		frstor [eax]
		add    eax, 4*8+80
		
		//4 DWORD segment registers
		xor  edx, edx
		
		mov  dx, gs
		mov  [eax], edx
		add  eax, 4
		
		mov  dx, fs
		mov  [eax], edx
		add  eax, 4

		mov  dx, es
		mov  [eax], edx
		add  eax, 4

		mov  dx, ds
		mov  [eax], edx
		add  eax, 4
		
		//6 DWORD integer registers
		mov [eax], edi
		add eax, 4
		
		mov [eax], esi
		add eax, 4
		
		mov [eax], ebx
		add eax, 4
		
		//Edx
		mov edx, [esp + 4]
		mov [eax], edx
		add eax, 4
		
		mov [eax], ecx
		add eax, 4
		
		//Eax
		mov edx, [esp + 8]
		mov [eax], edx
		add eax, 4
		
		//6 DWORD control registers
		mov [eax], ebp
		add eax, 4
		
		//Eip
		mov edx, [esp + 12]
		mov [eax], edx
		add  eax, 4
		
		xor  edx, edx
		mov  dx, cs
		mov  [eax], edx
		add  eax, 4
		
		//eflags
		mov edx, [esp]
		mov [eax], edx
		add eax, 4
		
		mov edx, esp
		add edx, 16
		mov [eax], edx
		add eax, 4
		
		xor  edx, edx
		mov  dx, ss
		mov  [eax], edx
		add  eax, 4		
		
		add esp, 12
		ret 4
	};
}


/*
77f91bb8 !
........ ! ;********************************************************
........ ! ; exported function KiUserExceptionDispatcher, ordinal 0021
........ ! ;********************************************************
........ ! exp_KiUserExceptionDispatcher_21:
........ ! 8b4c2404                         mov         ecx, [esp+4]
77f91bbc ! 8b1c24                           mov         ebx, [esp]
77f91bbf ! 51                               push        ecx
77f91bc0 ! 53                               push        ebx
77f91bc1 ! e8ecaf0100                       call        sub_77facbb2
77f91bc6 ! 0ac0                             or          al, al
77f91bc8 ! 740c                             jz          loc_77f91bd6

77facbb2 !
........ ! ;-----------------------
........ ! ;  S U B R O U T I N E
........ ! ;-----------------------
........ ! sub_77facbb2:                   ;xref c77f91bc1
........ ! 55                               push        ebp
77facbb3 ! 8bec                             mov         ebp, esp
77facbb5 ! 83ec60                           sub         esp, 60h
77facbb8 ! 53                               push        ebx
77facbb9 ! 56                               push        esi
*/

int __declspec(naked) WINAPI 
  RtlDispatchException(PEXCEPTION_RECORD pExcptRec, CONTEXT * pContext) {
	__asm {
		push ebp
		mov  ebp, esp
		sub  esp, 60h
		jmp dword ptr RtlDispatchException_resume
	};
}

int WINAPI VectoredExceptionDispatcher(PEXCEPTION_RECORD pExcptRec, CONTEXT * pContext) {
	ExceptionVectorList_t *list = &ExceptionVectorList;
	ExceptionVector_t *node;
	int ret;
	
	RtlEnterCriticalSection(&list->lock);
	node = list->head;
	while(node) {
		ret = node->handler(&pExcptRec);
		if(ret == EXCEPTION_CONTINUE_EXECUTION) return 1;
		node = node->next;
	}
	RtlLeaveCriticalSection(&list->lock);
	
	return RtlDispatchException(pExcptRec, pContext);
}

void InstallVectoredExceptionDispatcher(void) {
	//Check for KiUserExceptionDispatcher
	static const char signature[] =  { 0x8b, 0x4c, 0x24, 0x04, 0x8b, 0x1c, 0x24, 0x51, 0x53, 0xe8 };
	//Check for RtlDispatchException
	static const char signature2[] = { 0x55, 0x8b, 0xec, 0x83, 0xec, 0x60 };
	//For hook in RtlDispatchException
	static DWORD VectoredHandlerAddress = (DWORD)VectoredExceptionDispatcher;

	PBYTE KiUserExceptionDispatcher;
	DWORD address, written;
	char patch[8];
	
	KiUserExceptionDispatcher = (PBYTE)GetProcAddress(GetModuleHandleA("ntdll.dll"), "KiUserExceptionDispatcher");
	if(!KiUserExceptionDispatcher) {
		DbgPrintf(DBG_ERROR, S_RED "kernel32: Failed to find KiUserExceptionDispatcher in ntdll\n");
		return;
	}
	
	/*Ok lets do a quick check incase of changes to the code*/
	if(RtlCompareMemory(KiUserExceptionDispatcher, signature, sizeof(signature)) != sizeof(signature)) {
		DbgPrintf(DBG_ERROR, S_RED "kernel32: Failed to verify KiUserExceptionDispatcher\n");
		return;
	}
	
	address = *(DWORD *)(KiUserExceptionDispatcher + sizeof(signature));
	//eip is KiUserExceptionDispatcher + sizeof(signature) + sizeof(DWORD) and
	//RtlDispatchException is address + eip
	address += (DWORD)(KiUserExceptionDispatcher + sizeof(signature) + sizeof(DWORD));
	
	/*Ok lets do a quick check incase of changes to the code*/
	if(RtlCompareMemory((PVOID)address, signature2, sizeof(signature2)) != sizeof(signature2)) {
		DbgPrintf(DBG_ERROR, S_RED "kernel32: Failed to verify RtlDispatchException\n");
		return;
	}
	
	//Initialize Vector list
	ExceptionVectorList.head = NULL;
	ExceptionVectorList.tail = NULL;
	RtlInitializeCriticalSection(&ExceptionVectorList.lock);
	
	//Ok lets setup a jump to our handler, 6 bytes fits just right
	//jmp VectoredExceptionDispatcher = ff 25 dword_address
	RtlDispatchException_resume = address + 6;
	patch[0] = 0xff; patch[1]=0x25; *(DWORD *)&patch[2] = (DWORD)&VectoredHandlerAddress;
	if(!WriteProcessMemory(GetCurrentProcess(), (PVOID)address, patch, 6, &written)) {
		DbgPrintf(DBG_ERROR, S_RED "kernel32: Failed to patch RtlDispatchException");
		return;
	}
}

PVOID WINAPI XP_AddVectoredExceptionHandler(ULONG FirstHandler, PVECTORED_EXCEPTION_HANDLER VectoredHandler) {
	ExceptionVectorList_t *list = &ExceptionVectorList;
	ExceptionVector_t *new_node;
	
	if(!VectoredHandler) return NULL;
	
	new_node = (ExceptionVector_t *)malloc(sizeof(ExceptionVector_t));
	new_node->handler = VectoredHandler;
	
	RtlEnterCriticalSection(&list->lock);
	
	if(FirstHandler) {
		new_node->next = list->head;
		list->head = new_node;
		if(!list->tail) list->tail = new_node;
	} else {
		new_node->next = NULL;
		if(list->tail) list->tail->next = new_node;
		list->tail = new_node;
		if(!list->head) list->head = new_node;
	}
	
	RtlLeaveCriticalSection(&list->lock);
	
	return VectoredHandler;
}

ULONG WINAPI XP_RemoveVectoredExceptionHandler(PVOID Handler) {
	ExceptionVectorList_t *list = &ExceptionVectorList;
	ExceptionVector_t *node;
	ExceptionVector_t *node_prev;
	
	RtlEnterCriticalSection(&list->lock);
	node = list->head;
	node_prev = NULL;
	while(node) {
		if(node->handler == Handler) {
			if(node_prev) node_prev->next = node->next;
			if(list->head == node) list->head = node->next;
			if(list->tail == node) list->tail = node_prev;
			free(node);
			RtlLeaveCriticalSection(&list->lock);
			return 1;
		}
		node_prev = node;
		node = node->next;
	}
	RtlLeaveCriticalSection(&list->lock);
	
	return 0;
}

BOOL WINAPI XP_TzSpecificLocalTimeToSystemTime(
  LPTIME_ZONE_INFORMATION lpTimeZoneInformation,
  LPSYSTEMTIME lpLocalTime,
  LPSYSTEMTIME lpUniversalTime) {
	SYSTEMTIME temp;
	__int64 time1, time2;
	if(SystemTimeToTzSpecificLocalTime(lpTimeZoneInformation, lpLocalTime, &temp)) {
		//We convert the LocalTime as SystemTime to LocalTime, so we just need to inverse the op
		SystemTimeToFileTime(&temp, (PFILETIME)&time1);
		SystemTimeToFileTime(lpLocalTime, (PFILETIME)&time2);
		time1 = (time2 + time2 - time1);
		FileTimeToSystemTime((PFILETIME)&time1, lpUniversalTime);
		return TRUE;
	}
	return FALSE;
}

typedef HANDLE HACTCTX;

HANDLE WINAPI XP_CreateActCtxA(HACTCTX pActCtx) {
	DbgPrintf(DBG_WARN, S_YELLOW "kernel32: CreateActCtxW source='%s' flags=0x%03X stub\n", *(PDWORD)((PBYTE)pActCtx + 8), *(PDWORD)((PBYTE)pActCtx + 4));
	return pActCtx;
}

HANDLE WINAPI XP_CreateActCtxW(HACTCTX pActCtx) {
	DbgPrintf(DBG_WARN, S_YELLOW "kernel32: CreateActCtxW source='%S' flags=0x%03X stub\n", *(PDWORD)((PBYTE)pActCtx + 8), *(PDWORD)((PBYTE)pActCtx + 4));
	return pActCtx;
}

BOOL WINAPI XP_ActivateActCtx(HACTCTX hActCtx, ULONG_PTR* lpCookie) {
	DbgPrintf(DBG_WARN, S_YELLOW "kernel32: ActivateActCtx stub hActCtx=%08X\n", hActCtx);
	return TRUE;
}

BOOL WINAPI XP_DeactivateActCtx(DWORD dwFlags, ULONG_PTR ulCookie) {
	DbgPrintf(DBG_WARN, S_YELLOW "kernel32: DeactivateActCtx stub flags=%08X\n", dwFlags);
	return TRUE;
}

void WINAPI XP_ReleaseActCtx(HANDLE hActCtx) {
	DbgPrintf(DBG_WARN, S_YELLOW "kernel32: ReleaseActCtx stub hActCtx=%08X\n", hActCtx);
}

BOOL WINAPI XP_GetCurrentActCtx(HANDLE* lphActCtx) {
	DbgPrintf(DBG_WARN, S_YELLOW "kernel32: GetCurrentActCtx stub\n");
	return TRUE;
}

void WINAPI XP_AddRefActCtx(HANDLE hActCtx) {
	DbgPrintf(DBG_WARN, S_YELLOW "kernel32: AddRefActCtx stub\n");
}

BOOL WINAPI XP_ZombifyActCtx(HANDLE hActCtx) {
	DbgPrintf(DBG_WARN, S_YELLOW "kernel32: ZombifyActCtx stub\n");
	return TRUE;
}

BOOL WINAPI XP_QueryActCtxW(
	DWORD dwFlags, HANDLE hActCtx, PVOID pvSubInstance, ULONG ulInfoClass,
	PVOID pvBuffer, SIZE_T cbBuffer, SIZE_T *pcbWrittenOrRequired) {
	DbgPrintf(DBG_WARN, S_YELLOW "kernel32: QueryActCtxW stub\n");
	return FALSE;
}

BOOL WINAPI XP_FindActCtxSectionStringW(
	DWORD dwFlags, const GUID* lpExtensionGuid, ULONG ulSectionId,
	PWSTR lpStringToFind, PVOID ReturnedData) {
	DbgPrintf(DBG_WARN, S_YELLOW 
		"kernel32: FindActCtxSectionStringW flags=0x%08X id=%d string=%S\n",
		dwFlags, ulSectionId, lpStringToFind);
	return TRUE;  //must return true for MSVCR80
}

BOOL WINAPI XP_FindActCtxSectionStringA(
	DWORD dwFlags, const GUID* lpExtensionGuid, ULONG ulSectionId,
	PSTR lpStringToFind, PVOID ReturnedData) {
	DbgPrintf(DBG_WARN, S_YELLOW 
		"kernel32: FindActCtxSectionStringA flags=0x%08X id=%d string=%s\n",
		dwFlags, ulSectionId, lpStringToFind);
	return TRUE;  //must return true for MSVCR80
}

BOOL __declspec(naked) SSE2_Available() {
  __asm {
	push ecx
	push ebx
	mov eax, 1
	cpuid
	mov eax, BIT(26)
	and eax, edx
	shr eax, 26
	pop ebx
	pop ecx
	ret
  }
}

//FIXME: add option for SSE2 faking
BOOL WINAPI XP_IsProcessorFeaturePresent(DWORD ProcessorFeature) {
	if(ProcessorFeature == PF_XMMI64_INSTRUCTIONS_AVAILABLE) return SSE2_Available();
	return IsProcessorFeaturePresent(ProcessorFeature);
}

BOOL WINAPI XP_GetDiskFreeSpaceA(
  LPCSTR lpRootPathName,
  LPDWORD lpSectorsPerCluster,
  LPDWORD lpBytesPerSector,
  LPDWORD lpNumberOfFreeClusters,
  LPDWORD lpTotalNumberOfClusters) {
	if(!GetDiskFreeSpaceA(lpRootPathName, lpSectorsPerCluster, lpBytesPerSector, lpNumberOfFreeClusters, lpTotalNumberOfClusters)
	   && GetLastError() == ERROR_DIR_NOT_ROOT)
	   return GetDiskFreeSpaceW(NULL, lpSectorsPerCluster, lpBytesPerSector, lpNumberOfFreeClusters, lpTotalNumberOfClusters);
	return FALSE;
}

BOOL WINAPI XP_GetDiskFreeSpaceW(
  LPCWSTR lpRootPathName,
  LPDWORD lpSectorsPerCluster,
  LPDWORD lpBytesPerSector,
  LPDWORD lpNumberOfFreeClusters,
  LPDWORD lpTotalNumberOfClusters) {
	if(!GetDiskFreeSpaceW(lpRootPathName, lpSectorsPerCluster, lpBytesPerSector, lpNumberOfFreeClusters, lpTotalNumberOfClusters)
	   && GetLastError() == ERROR_DIR_NOT_ROOT)
	   return GetDiskFreeSpaceW(NULL, lpSectorsPerCluster, lpBytesPerSector, lpNumberOfFreeClusters, lpTotalNumberOfClusters);
	return FALSE;
}

BOOL WINAPI XP_IsProcessInJob(HANDLE ProcessHandle, HANDLE JobHandle, PBOOL Result) {
	char buffer[sizeof(JOBOBJECT_BASIC_PROCESS_ID_LIST) + sizeof(DWORD)*64];
	DWORD pid, i;
	PJOBOBJECT_BASIC_PROCESS_ID_LIST pid_list;
	
	pid_list = (PJOBOBJECT_BASIC_PROCESS_ID_LIST)buffer;
	pid_list->NumberOfAssignedProcesses = pid_list->NumberOfProcessIdsInList = 64;
	
	if(!Result) return FALSE;
	
	pid = XP_GetProcessId(ProcessHandle);
	
	if(JobHandle) {
		if(QueryInformationJobObject(JobHandle, JobObjectBasicProcessIdList, buffer, sizeof(buffer), NULL)) {
			for(i = 0; i < pid_list->NumberOfProcessIdsInList; i++) {
				if(pid_list->ProcessIdList[i] == pid) { *Result = TRUE; return TRUE; }
			}
			*Result = FALSE;
			return TRUE;
		} else {
			return FALSE;
		}
	} else {
		/*                        FIXME!!!!!
		  This info is stored in the EPROCESS structure which you needs the
		  kernel mode for.  Dunno any other way to do this one.*/
		return FALSE;
	}
}

LANGID WINAPI XP_SetThreadUILanguage(LANGID LangId) {
	LCID lcid;
	if (LangId != 0)
		return 0;
	
	lcid = GetUserDefaultLCID();
	SetThreadLocale(lcid);
	return LANGIDFROMLCID(lcid);
}

BOOL WINAPI XP_SetFileValidData(HANDLE hFile, LONGLONG ValidDataLength) {
	LARGE_INTEGER currFilePos, setFilePos;
	BOOL setOkay;
	
	setFilePos.QuadPart = 0ll;
	if(!SetFilePointerEx(hFile, setFilePos, &currFilePos, FILE_CURRENT))
		return FALSE;
	
	setFilePos.QuadPart = ValidDataLength;
	if(SetFilePointerEx(hFile, setFilePos, NULL, FILE_BEGIN)) {
		setOkay = SetEndOfFile(hFile);
		return SetFilePointerEx(hFile, currFilePos, NULL, FILE_BEGIN) && setOkay;
	}
	
	return FALSE;
}

BOOL WINAPI XP_GetLogicalProcessorInformation(PSYSTEM_LOGICAL_PROCESSOR_INFORMATION Buffer, PDWORD ReturnLength) {
	if(*ReturnLength < sizeof(SYSTEM_LOGICAL_PROCESSOR_INFORMATION)) {
		SetLastError(ERROR_INSUFFICIENT_BUFFER);
		*ReturnLength = sizeof(SYSTEM_LOGICAL_PROCESSOR_INFORMATION);
		return FALSE;
	}
	
	DbgPrintf(DBG_WARN, S_YELLOW "kernel32: GetLogicalProcessorInformation mask:%x relation:%x %llx %llx\n",
		Buffer->ProcessorMask,
		Buffer->Relationship,
		Buffer->Reserved[0], Buffer->Reserved[1]);
	
	SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
	return FALSE;
}

BOOL WINAPI XP_SetProcessDEPPolicy(DWORD dwFlags) {
	DbgPrintf(DBG_INFO, "kernel32: SetProcessDEPPolicy(%x)\n", dwFlags);
	return TRUE;
}

USHORT WINAPI XP_CaptureStackBackTrace(ULONG FramesToSkip, ULONG FramesToCapture, PVOID *BackTrace, PULONG BackTraceHash) {
	DbgPrintf(DBG_INFO, "kernel32: CaptureStackBackTrace not implemented\n");
	return 0;
}

PVOID WINAPI RtlProcessHeap();

/*
7c59c75b 64a118000000    mov     eax,dword ptr fs:[00000018h]
7c59c761 8b4030          mov     eax,dword ptr [eax+30h]
7c59c764 68e4020000      push    2E4h
7c59c769 ff357cc55c7c    push    dword ptr [KERNEL32!BaseDllTag (7c5cc57c)]
7c59c76f ff7018          push    dword ptr [eax+18h]
7c59c772 ff150410577c    call    dword ptr [KERNEL32!_imp__RtlAllocateHeap (7c571004)]
*/

BOOL WINAPI XP_ConvertFiberToThread() {
	PVOID Fiber;
	PTEB_ Teb;
	
	Teb = (PTEB_)NtCurrentTeb();
	Fiber = Teb->NtTib.FiberData;
	
	if (!Fiber) {
		return FALSE;
	}
	
	Teb->NtTib.FiberData = NULL;
	RtlFreeHeap(RtlProcessHeap(), 0, Fiber);
	
	return TRUE;
}

BOOL WINAPI DllMain(
  HINSTANCE hinstDLL,
  DWORD fdwReason,
  LPVOID lpvReserved) {
	static CHAR  ini_filename[] = INI_FILENAME;  
	static CHAR  buffer [MAX_PATH];
	static WCHAR bufferW[MAX_PATH];
	DWORD read;
	char *temp, *p;
	
	switch(fdwReason) {
		case DLL_PROCESS_ATTACH:
			thisModule   = hinstDLL;
			hookedModule = GetModuleHandleA("KERNEL32_ORG.DLL");
			
			DetourFunction((PBYTE)GetProcAddress(hookedModule, "LoadLibraryExW"), (PBYTE)XP_LoadLibraryExW);
			__asm mov _LoadLibraryExW, eax
			
			//DetourFunction((PBYTE)GETLOCALECOMPARE_ADDY, XP_GetLocaleCompare);
			//__asm mov GetLocaleCompare, eax
			
			versionInfoA.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEXA);
			versionInfoW.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEXW);
			GetVersionExA((LPOSVERSIONINFOA)&versionInfoA);
			GetVersionExW((LPOSVERSIONINFOW)&versionInfoW);
				
			GetSystemDirectoryW(systemDirectoryW, sizeof(systemDirectoryW));
			GetSystemDirectoryA(systemDirectoryA, sizeof(systemDirectoryA));

			read = GetCurrentDirectoryA(sizeof(buffer), buffer);
			if(read == 0 || read + sizeof(ini_filename) >= sizeof(buffer)) break;
			lstrcatA(buffer, ini_filename);
			
			MultiByteToWideChar(CP_ACP, 0, buffer, -1, bufferW, ARRAY_SIZE(bufferW));
			
			debug = GetPrivateProfileIntA(KERNEL32_SECTION, "debug", 0, buffer);
			if(DebugLevel()) InitDebug();
			
			DbgPrintf(DBG_TRACE, "kernel32: INI file path: %s\n", buffer);
			
			verifyVersion =
				GetPrivateProfileIntA(KERNEL32_SECTION, "verifyVersion", 0, buffer);
				
			versionInfoA.dwMajorVersion = versionInfoW.dwMajorVersion = 
				GetPrivateProfileIntA(KERNEL32_SECTION, "majVer", versionInfoA.dwMajorVersion, buffer);
			versionInfoA.dwMinorVersion = versionInfoW.dwMinorVersion = 
				GetPrivateProfileIntA(KERNEL32_SECTION, "minVer", versionInfoA.dwMinorVersion, buffer);
			versionInfoA.dwBuildNumber = versionInfoW.dwBuildNumber = 
				GetPrivateProfileIntA(KERNEL32_SECTION, "buildNum", versionInfoA.dwBuildNumber, buffer);
				
			versionInfoA.wServicePackMajor = versionInfoW.wServicePackMajor = 
				GetPrivateProfileIntA(KERNEL32_SECTION, "majSP", versionInfoA.wServicePackMajor, buffer);
			versionInfoA.wServicePackMinor = versionInfoW.wServicePackMinor = 
				GetPrivateProfileIntA(KERNEL32_SECTION, "minSP", versionInfoA.wServicePackMinor, buffer);
							
			GetPrivateProfileStringA(KERNEL32_SECTION, "versionString", versionInfoA.szCSDVersion, versionInfoA.szCSDVersion, sizeof(versionInfoA.szCSDVersion), buffer);
			MultiByteToWideChar(CP_ACP, 0, versionInfoA.szCSDVersion, -1, versionInfoW.szCSDVersion, ARRAY_SIZE(versionInfoW.szCSDVersion));
			
			versionInfoA.wSuiteMask = versionInfoW.wSuiteMask= VER_SUITE_TERMINAL|VER_SUITE_SINGLEUSERTS;
			versionInfoA.wProductType = versionInfoW.wProductType = VER_NT_WORKSTATION;
			
			DbgPrintf(DBG_INFO, "kernel32: majVer=%d minVer=%d buildNum=%d majSP=%hd minSP=%hd versionString=%s\n",
				versionInfoA.dwMajorVersion, versionInfoA.dwMinorVersion, versionInfoA.dwBuildNumber,
				versionInfoA.wServicePackMajor, versionInfoA.wServicePackMinor, versionInfoA.szCSDVersion);
			
			/*Version Faking Filter*/
			temp = (char *)HeapAlloc(GetProcessHeap(), 0, sizeof(CHAR) * VERSION_MOD_LIST_SIZE);
			RtlZeroMemory(filterModuleNames, sizeof(filterModuleNames));
			RtlZeroMemory(filterModules, sizeof(filterModules));
			
			GetPrivateProfileStringA(KERNEL32_SECTION, "versionFaking", S_VERSION_FAKING_EXCLUDE, temp, sizeof(CHAR) * VERSION_MOD_LIST_SIZE, buffer);
			
			if(lstrcmpiA(temp, S_VERSION_FAKING_INCLUDE) == 0) versionFaking = VERSION_FAKING_INCLUDE;
			else                                               versionFaking = VERSION_FAKING_EXCLUDE;
			
			if(versionFaking == VERSION_FAKING_INCLUDE)
				DbgPrintf(DBG_INFO, "kernel32: version faking for only the following modules\n");
			else
				DbgPrintf(DBG_INFO, "kernel32: version faking for all modules except the following\n");
			
			numFilterModules = 0;
			GetPrivateProfileSectionA(VERSION_SECTION, temp, sizeof(CHAR) * VERSION_MOD_LIST_SIZE, buffer);
			for(p = temp; *p; ) {
				*strchr(p, '=') = '\0';
				filterModuleNames[numFilterModules] = p;
				filterModules[numFilterModules++]   = XP_GetModuleHandleA(p);
				DbgPrintf(DBG_INFO, "kernel32:    %s\n", p);
				while(*p) p++;    //skip over key
				p++;              //the '=' which has been nulled
				while(*p) p++;    //skip over value
				p++;
			}
			
			GetModuleOriginal = GetPrivateProfileIntA(KERNEL32_SECTION, "GetModuleOriginal", 0, buffer);
			
			//Default to United States
			geoNation = GetPrivateProfileIntA(KERNEL32_SECTION, "geoNation", 0xF4, buffer);
			geoRegion = GetPrivateProfileIntA(KERNEL32_SECTION, "geoRegion", 0xF4, buffer);
			
			GetPrivateProfileStringW(KERNEL32_SECTIONW, L"geoISO2",     L"US",  geoISO2W, ARRAY_SIZE(geoISO2W), bufferW);
			GetPrivateProfileStringW(KERNEL32_SECTIONW, L"geoISO3",     L"USA", geoISO3W, ARRAY_SIZE(geoISO3W), bufferW);
			GetPrivateProfileStringW(KERNEL32_SECTIONW, L"geoRFC1766",  L"USA", geoRFC1766W, ARRAY_SIZE(geoRFC1766W), bufferW);
			GetPrivateProfileStringW(KERNEL32_SECTIONW, L"geoFriendly", L"United States", geoFriendlyW, ARRAY_SIZE(geoFriendlyW), bufferW);
			GetPrivateProfileStringW(KERNEL32_SECTIONW, L"geoOfficial", L"United States of America", geoOfficialW, ARRAY_SIZE(geoOfficialW), bufferW);
			
			GetPrivateProfileStringA(KERNEL32_SECTION, "SystemDirectory", NULL, systemDirectoryA, sizeof(systemDirectoryA), buffer);
			if(systemDirectoryA[0] ==  '\0') GetSystemDirectoryA(systemDirectoryA, sizeof(systemDirectoryA));
			MultiByteToWideChar(CP_ACP, 0, systemDirectoryA, -1, systemDirectoryW, ARRAY_SIZE(systemDirectoryW));
			
			if(GetPrivateProfileIntA(KERNEL32_SECTION, "Load_ACLayers", 0, buffer)) {
				//oh noes I'm calling loadlibrary in dllmain
				if(!LoadLibraryA("aclayers.dll")) DbgPrintf(DBG_ERROR, S_RED "kernel32: Failed to load aclayers.dll!\n");
			}
			
			RewriteExportTable = GetPrivateProfileIntA(KERNEL32_SECTION, "RewriteExportTable", 0, buffer);
			if(RewriteExportTable) {
				RenameExportTableName(hookedModule, "KERNEL32_ORG.dll");
				ForwardExportsToFunctionExports(thisModule);
			}
			
			InstallVectoredExceptionDispatcher();

			RtlInitializeCriticalSection(&dllDirectoryLock);
			dllDirectoryW[0] = L'\0';
			break;
		case DLL_PROCESS_DETACH:
			break;
	}

	return TRUE;
}

